# Copyright (C) 2018 Cancer Care Associates

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.


"""Index logfiles.
"""

import os
import json
import pathlib
import traceback
from glob import glob

import attr

from pymedphys_utilities.filehash import hash_file
from pymedphys_utilities.utilities import get_sql_servers, make_a_valid_directory_name
from pymedphys_databases.msq import (
    multi_mosaiq_connect, get_mosaiq_delivery_details, OISDeliveryDetails,
    NoMosaiqEntries)
from pymedphys_fileformats.trf import Header, decode_header_from_file

from .identify import date_convert


def create_logfile_directory_name(centre,
                                  delivery_details: OISDeliveryDetails,
                                  header: Header, path_string_time):
    proposed_patient_directory_name = '{}_{}_{}'.format(
        delivery_details.patient_id, delivery_details.last_name,
        delivery_details.first_name
    )

    patient_directory_name = make_a_valid_directory_name(
        proposed_patient_directory_name)

    if delivery_details.qa_mode:
        clinical_or_qa = 'qa'
    else:
        clinical_or_qa = 'clinical'

    proposed_field_directory_name = '{}_{}_{}_{}'.format(
        delivery_details.field_id, header.field_label, header.field_name,
        delivery_details.field_type
    )

    field_directory_name = make_a_valid_directory_name(
        proposed_field_directory_name)

    proposed_delivery_directory = '{}_{}'.format(
        path_string_time, header.machine
    )

    delivery_directory = make_a_valid_directory_name(
        proposed_delivery_directory)

    return os.path.join(
        centre, patient_directory_name, clinical_or_qa,
        field_directory_name, delivery_directory
    )


def create_index_entry(new_filepath, delivery_details: OISDeliveryDetails,
                       header: Header, mosaiq_string_time):

    index_entry = {
        'filepath': new_filepath,
        'delivery_details': {
            **attr.asdict(delivery_details)
        },
        'logfile_header': {
            **header._asdict()
        },
        'local_time': mosaiq_string_time
    }

    return index_entry


def file_already_in_index(indexed_filepath, to_be_indexed_filepath, filehash):
    try:
        new_hash = hash_file(indexed_filepath)
    except FileNotFoundError:
        raise FileNotFoundError(
            "Indexed logfile can't be found in its declared location.")

    try:
        assert new_hash == filehash
    except AssertionError:
        raise AssertionError(
            "The located file doesn't agree with the index hash.")

    print('Already exists in index')
    os.remove(to_be_indexed_filepath)


def rename_and_handle_fileexists(old_filepath, new_filepath):
    try:
        os.rename(old_filepath, new_filepath)
    except FileExistsError:
        files_hashes_match = (
            hash_file(old_filepath) ==
            hash_file(new_filepath)
        )
        print('    File already exists at {}'.format(new_filepath))
        if files_hashes_match:
            os.remove(old_filepath)
        else:
            raise FileExistsError(
                "File already exists and the hash does not match")


# TODO Split this function up into smaller functions for easier reuse.
def file_ready_to_be_indexed(cursors, filehash_list, to_be_indexed_dict,
                             unknown_error_in_logfile, no_mosaiq_record_found,
                             no_field_label_in_logfile,
                             indexed_directory, index_filepath, index,
                             machine_map, centre_details, centre_server_map):
    for filehash in filehash_list:
        logfile_basename = os.path.basename(to_be_indexed_dict[filehash])

        try:
            header = decode_header_from_file(to_be_indexed_dict[filehash])
            print("\n{}".format(header))
            if header.field_label == "":
                print('No field label in logfile')
                new_filepath = os.path.join(
                    no_field_label_in_logfile, logfile_basename)
                rename_and_handle_fileexists(
                    to_be_indexed_dict[filehash], new_filepath)
                continue

            centre = machine_map[header.machine]['centre']
            server = centre_server_map[centre]

            mosaiq_string_time, path_string_time = date_convert(
                header.date, centre_details[centre]['timezone'])
        except Exception as e:
            traceback.print_exc()
            new_filepath = os.path.join(
                unknown_error_in_logfile, logfile_basename)
            rename_and_handle_fileexists(
                to_be_indexed_dict[filehash], new_filepath)
            continue

        try:
            delivery_details = get_mosaiq_delivery_details(
                cursors[server], header.machine, mosaiq_string_time,
                header.field_label, header.field_name, buffer=240)
        except NoMosaiqEntries as e:
            print(e)
            new_filepath = os.path.join(
                no_mosaiq_record_found, logfile_basename)
            rename_and_handle_fileexists(
                to_be_indexed_dict[filehash], new_filepath)
            continue

        logfile_directory_name = create_logfile_directory_name(
            centre, delivery_details, header,
            path_string_time)

        abs_logfile_directory_name = os.path.abspath(
            os.path.join(indexed_directory, logfile_directory_name))

        pathlib.Path(abs_logfile_directory_name).mkdir(
            parents=True, exist_ok=True)

        new_filepath = os.path.join(
            logfile_directory_name,
            logfile_basename)

        index[filehash] = create_index_entry(
            new_filepath, delivery_details, header, mosaiq_string_time)

        abs_new_filepath = os.path.abspath(
            os.path.join(indexed_directory, new_filepath))

        index_extension_split = os.path.splitext(index_filepath)
        temp_index_path = "{}_temp{}".format(
            index_extension_split[0], index_extension_split[1])

        with open(temp_index_path, 'w') as json_data_file:
            json.dump(index, json_data_file, indent=2)

        os.replace(temp_index_path, index_filepath)

        os.rename(to_be_indexed_dict[filehash], abs_new_filepath)

        print('Indexed logfile:\n    {} -->\n    {}'.format(
            to_be_indexed_dict[filehash], abs_new_filepath))


def index_logfiles(centre_map, machine_map, logfile_data_directory):
    data_directory = logfile_data_directory
    index_filepath = os.path.abspath(
        os.path.join(data_directory, 'index.json'))
    to_be_indexed_directory = os.path.abspath(
        os.path.join(data_directory, 'to_be_indexed'))
    indexed_directory = os.path.abspath(
        os.path.join(data_directory, 'indexed'))
    no_mosaiq_record_found = os.path.abspath(
        os.path.join(data_directory, 'no_mosaiq_record_found'))
    unknown_error_in_logfile = os.path.abspath(
        os.path.join(data_directory, 'unknown_error_in_logfile'))
    no_field_label_in_logfile = os.path.abspath(
        os.path.join(data_directory, 'no_field_label_in_logfile'))
    # machine_map = config['machine_map']
    centre_details = centre_map

    centre_server_map = {
        centre: centre_lookup['mosaiq_sql_server']
        for centre, centre_lookup in centre_map.items()
    }

    sql_server_and_ports = [
        "{}".format(details['mosaiq_sql_server'])
        for _, details in centre_details.items()
    ]

    with open(index_filepath, 'r') as json_data_file:
        index = json.load(json_data_file)

    indexset = set(index.keys())

    print('\nConnecting to Mosaiq SQL servers...')
    with multi_mosaiq_connect(sql_server_and_ports) as cursors:

        print('Globbing index directory...')
        to_be_indexed = glob(
            os.path.join(to_be_indexed_directory, '**/*.trf'), recursive=True)

        chunk_size = 50
        number_to_be_indexed = len(to_be_indexed)
        to_be_indexed_chunked = [
            to_be_indexed[i:i+chunk_size]
            for i in range(0, number_to_be_indexed, chunk_size)]

        for i, a_to_be_indexed_chunk in enumerate(to_be_indexed_chunked):
            print('\nHashing a chunk of logfiles ({}/{})'.format(
                i + 1, len(to_be_indexed_chunked)))
            hashlist = [
                hash_file(filename, dot_feedback=True)
                for filename in a_to_be_indexed_chunk
            ]

            print(' ')

            to_be_indexed_dict = dict(zip(hashlist, a_to_be_indexed_chunk))

            hashset = set(hashlist)

            for filehash in list(hashset.intersection(indexset)):
                file_already_in_index(
                    os.path.join(
                        indexed_directory, index[filehash]['filepath']),
                    to_be_indexed_dict[filehash],
                    filehash
                )

            file_ready_to_be_indexed(
                cursors, list(hashset.difference(indexset)),
                to_be_indexed_dict,
                unknown_error_in_logfile, no_mosaiq_record_found,
                no_field_label_in_logfile,
                indexed_directory, index_filepath, index,
                machine_map, centre_details, centre_server_map
            )
    print('Complete')



# Copyright (C) 2018 Cancer Care Associates

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.


"""Analyse logfiles.
"""

import os
import traceback
import json

import numpy as np
import matplotlib.pyplot as plt

from pymedphys_utilities.utilities import (
    get_cache_filepaths, get_mu_density_parameters,
    get_index, get_centre, get_sql_servers, get_sql_servers_list,
    get_filepath
)
from pymedphys_databases.msq import multi_mosaiq_connect, multi_fetch_and_verify_mosaiq
from pymedphys_databases.delivery import DeliveryDatabases
from pymedphys_mudensity.mudensity import calc_mu_density


def analyse_single_hash(index, config, filehash, cursors):
    field_id_key_map = get_field_id_key_map(index)
    logfile_filepath = get_filepath(index, config, filehash)
    print(logfile_filepath)

    results = get_logfile_mosaiq_results(
        index, config, logfile_filepath, field_id_key_map,
        filehash, cursors,
        grid_resolution=5/3)

    comparison = calc_comparison(results[2], results[3])
    print("Comparison result = {}".format(comparison))
    plot_results(*results)

    return comparison


def load_comparisons_from_cache(config):
    (
        comparison_storage_filepath, _
    ) = get_cache_filepaths(config)

    with open(comparison_storage_filepath, 'r') as comparisons_file:
        comparison_storage = json.load(comparisons_file)

    file_hashes = np.array(list(comparison_storage.keys()))

    comparisons = np.array([
        comparison_storage[file_hash]
        for file_hash in file_hashes
    ])

    index = get_index(config)
    file_paths_worst_first = np.array([
        get_filepath_from_hash(config, index, file_hash)
        for file_hash in file_hashes
    ])

    sort_ref = np.argsort(comparisons)[::-1]

    file_hashes = file_hashes[sort_ref]
    # comparisons = comparisons[sort_ref]
    file_paths_worst_first = file_paths_worst_first[sort_ref]

    return file_hashes, comparison_storage, file_paths_worst_first


def get_field_id_key_map(index):
    field_id_key_map = dict()

    for key, value in index.items():
        if not value['delivery_details']['qa_mode']:
            field_id = value['delivery_details']['field_id']
            if field_id not in field_id_key_map:
                field_id_key_map[field_id] = []

            field_id_key_map[field_id].append(key)

    return field_id_key_map


def random_uncompared_logfiles(index, config, compared_hashes):
    index_set = set(index.keys())
    comparison_set = set(compared_hashes)

    not_yet_compared = np.array(list(index_set.difference(comparison_set)))
    field_types = np.array([
        index[file_hash]['delivery_details']['field_type']
        for file_hash in not_yet_compared
    ])

    file_hashes_vmat = not_yet_compared[field_types == 'VMAT']

    vmat_filepaths = np.array([
        os.path.join(
            config['linac_logfile_data_directory'],
            'indexed',
            index[file_hash]['filepath'])
        for file_hash in file_hashes_vmat
    ])

    shuffle_index = np.arange(len(vmat_filepaths))
    np.random.shuffle(shuffle_index)

    return file_hashes_vmat[shuffle_index], vmat_filepaths[shuffle_index]


def mudensity_comparisons(config, plot=True, new_logfiles=False):
    (
        comparison_storage_filepath, comparison_storage_scratch
    ) = get_cache_filepaths(config)

    grid_resolution, ram_fraction = get_mu_density_parameters(config)

    index = get_index(config)
    field_id_key_map = get_field_id_key_map(index)

    (
        file_hashes, comparisons, _
    ) = load_comparisons_from_cache(config)

    if new_logfiles:
        file_hashes, _ = random_uncompared_logfiles(
            index, config, file_hashes)

    sql_servers_list = get_sql_servers_list(config)

    with multi_mosaiq_connect(sql_servers_list) as cursors:
        for file_hash in file_hashes:

            try:
                logfile_filepath = get_filepath(index, config, file_hash)
                print("\n{}".format(logfile_filepath))

                if (new_logfiles) and (file_hash in comparisons):
                    raise AssertionError(
                        "A new logfile shouldn't have already been compared")

                if index[file_hash]['delivery_details']['qa_mode']:
                    print('Skipping QA field')
                else:
                    if file_hash in comparisons:
                        print("Cached comparison value = {}".format(
                            comparisons[file_hash]))

                    results = get_logfile_mosaiq_results(
                        index, config, logfile_filepath, field_id_key_map,
                        file_hash, cursors,
                        grid_resolution=grid_resolution)
                    new_comparison = calc_comparison(results[2], results[3])

                    if file_hash not in comparisons:
                        update_comparison_file(
                            file_hash, new_comparison,
                            comparison_storage_filepath,
                            comparison_storage_scratch)
                        print("Newly calculated comparison value = {}".format(
                            new_comparison))
                    elif np.abs(comparisons[file_hash] - new_comparison) > 0.00001:
                        print(
                            "Calced comparison value does not agree with the "
                            "cached value.")
                        print("Newly calculated comparison value = {}".format(
                            new_comparison))
                        update_comparison_file(
                            file_hash, new_comparison,
                            comparison_storage_filepath,
                            comparison_storage_scratch)
                        print("Overwrote the cache with the new result.")
                    else:
                        print(
                            "Calced comparison value agrees with the cached value")
                    if plot:
                        plot_results(*results)
            except KeyboardInterrupt:
                raise
            except AssertionError:
                raise
            except Exception:
                print(traceback.format_exc())


def mu_density_from_delivery_data(delivery_data, grid_resolution=1):
    mu, mlc, jaw = (
        delivery_data.monitor_units, delivery_data.mlc, delivery_data.jaw)

    grid_xx, grid_yy, mu_density = calc_mu_density(
        mu, mlc, jaw,
        grid_resolution=grid_resolution)

    return grid_xx, grid_yy, mu_density


def find_consecutive_logfiles(field_id_key_map, field_id, filehash, index,
                              config):
    keys = np.array(field_id_key_map[field_id])

    times = np.array([
        index[key]['local_time']
        for key in keys
    ]).astype(np.datetime64)

    sort_reference = np.argsort(times)
    keys = keys[sort_reference]
    times = times[sort_reference]

    hours_4 = np.array(60 * 60 * 4).astype(np.timedelta64)

    delivery_time = np.array(
        index[filehash]['local_time']).astype(np.datetime64)
    within_4_hours_reference = np.abs(delivery_time - times) < hours_4
    within_4_hours = keys[within_4_hours_reference].tolist()

    return within_4_hours


def calc_and_merge_logfile_mudensity(filepaths, grid_resolution=1):
    logfile_results = []
    for filepath in filepaths:
        logfile_delivery_data = DeliveryDatabases.from_logfile(filepath)
        mu_density_results = mu_density_from_delivery_data(
            logfile_delivery_data, grid_resolution=grid_resolution)

        logfile_results.append(mu_density_results)

    grid_xx_list = [
        result[0] for result in logfile_results
    ]
    grid_yy_list = [
        result[1] for result in logfile_results
    ]

    # assert np.array_equal(*grid_xx_list)
    # assert np.array_equal(*grid_yy_list)

    grid_xx = grid_xx_list[0]
    grid_yy = grid_yy_list[0]

    mu_densities = [
        result[2] for result in logfile_results
    ]

    logfile_mu_density = np.sum(mu_densities, axis=0)

    return grid_xx, grid_yy, logfile_mu_density


def get_logfile_mosaiq_results(index, config, filepath, field_id_key_map,
                               filehash, cursors, grid_resolution=1):
    file_info = index[filehash]
    delivery_details = file_info['delivery_details']
    field_id = delivery_details['field_id']

    centre = get_centre(config, file_info)
    server = get_sql_servers(config)[centre]
    mosaiq_delivery_data = multi_fetch_and_verify_mosaiq(
        cursors[server], field_id)

    mosaiq_results = mu_density_from_delivery_data(
        mosaiq_delivery_data, grid_resolution=grid_resolution)

    consecutive_keys = find_consecutive_logfiles(
        field_id_key_map, field_id, filehash, index, config)

    logfilepaths = [
        get_filepath(index, config, key)
        for key in consecutive_keys
    ]

    logile_results = calc_and_merge_logfile_mudensity(
        logfilepaths, grid_resolution=grid_resolution)

    try:
        assert np.all(logile_results[0] == mosaiq_results[0])
        assert np.all(logile_results[1] == mosaiq_results[1])
    except AssertionError:
        print(np.shape(logile_results[0]))
        print(np.shape(mosaiq_results[0]))
        raise

    grid_xx = logile_results[0]
    grid_yy = logile_results[1]

    logfile_mu_density = logile_results[2]
    mosaiq_mu_density = mosaiq_results[2]

    return grid_xx, grid_yy, logfile_mu_density, mosaiq_mu_density


def calc_comparison(logfile_mu_density, mosaiq_mu_density, normalisation=None):
    if normalisation is None:
        normalisation = np.sum(mosaiq_mu_density)

    comparison = (
        np.sum(np.abs(logfile_mu_density - mosaiq_mu_density)) / normalisation)

    return comparison


# def get_logfile_mosaiq_comparison(index, config, filepath, cursors,
#                                   grid_resolution=1, ram_fraction=0.8):
#     _, _, logfile_mu_density, mosaiq_mu_density = get_logfile_mosaiq_results(
#         index, config, filepath, cursors, grid_resolution=grid_resolution,
#         ram_fraction=ram_fraction)

#     return calc_comparison(logfile_mu_density, mosaiq_mu_density)


def get_filepath_from_hash(config, index, file_hash):
    return os.path.join(
        config['linac_logfile_data_directory'],
        'indexed',
        index[file_hash]['filepath'])


def plot_results(grid_xx, grid_yy, logfile_mu_density, mosaiq_mu_density,
                 diff_colour_scale=0.1):
    min_val = np.min([logfile_mu_density, mosaiq_mu_density])
    max_val = np.max([logfile_mu_density, mosaiq_mu_density])

    plt.figure()
    plt.pcolormesh(
        grid_xx,
        grid_yy,
        logfile_mu_density,
        vmin=min_val, vmax=max_val)
    plt.colorbar()
    plt.title('Logfile MU density')
    plt.xlabel('MLC direction (mm)')
    plt.ylabel('Jaw direction (mm)')
    plt.gca().invert_yaxis()

    plt.figure()
    plt.pcolormesh(
        grid_xx,
        grid_yy,
        mosaiq_mu_density,
        vmin=min_val, vmax=max_val)
    plt.colorbar()
    plt.title('Mosaiq MU density')
    plt.xlabel('MLC direction (mm)')
    plt.ylabel('Jaw direction (mm)')
    plt.gca().invert_yaxis()

    scaled_diff = (logfile_mu_density - mosaiq_mu_density) / max_val

    plt.figure()
    plt.pcolormesh(
        grid_xx,
        grid_yy,
        scaled_diff,
        vmin=-diff_colour_scale/2, vmax=diff_colour_scale/2)
    plt.colorbar(label='Limited colour range = {}'.format(diff_colour_scale/2))
    plt.title('(Logfile - Mosaiq MU density) / Maximum MU Density')
    plt.xlabel('MLC direction (mm)')
    plt.ylabel('Jaw direction (mm)')
    plt.gca().invert_yaxis()

    plt.show()

    plt.figure()
    plt.pcolormesh(
        grid_xx,
        grid_yy,
        scaled_diff,
        vmin=-diff_colour_scale, vmax=diff_colour_scale)
    plt.colorbar(label='Limited colour range = {}'.format(diff_colour_scale))
    plt.title('(Logfile - Mosaiq MU density) / Maximum MU Density')
    plt.xlabel('MLC direction (mm)')
    plt.ylabel('Jaw direction (mm)')
    plt.gca().invert_yaxis()

    plt.show()

    absolute_range = np.max([-np.min(scaled_diff), np.max(scaled_diff)])

    plt.figure()
    plt.pcolormesh(
        grid_xx,
        grid_yy,
        scaled_diff,
        vmin=-absolute_range, vmax=absolute_range)
    plt.colorbar(label='No limited colour range')
    plt.title('(Logfile - Mosaiq MU density) / Maximum MU Density')
    plt.xlabel('MLC direction (mm)')
    plt.ylabel('Jaw direction (mm)')
    plt.gca().invert_yaxis()

    plt.show()


def update_comparison_file(file_hash, comparison, comparison_storage_filepath,
                           comparison_storage_scratch):
    with open(comparison_storage_filepath, 'r') as comparisons_file:
        comparison_storage = json.load(comparisons_file)

    comparison_storage[file_hash] = comparison

    with open(comparison_storage_scratch, 'w') as comparisons_file:
        json.dump(comparison_storage, comparisons_file)

    os.replace(comparison_storage_scratch, comparison_storage_filepath)



# Copyright (C) 2018 Cancer Care Associates

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.


import numpy as np

from pymedphys_utilities.utilities import get_filepath, get_gantry_tolerance

from pymedphys_mudensity.mudensity import get_grid


from pymedphys_databases.msq import multi_fetch_and_verify_mosaiq
from pymedphys_databases.delivery import DeliveryDatabases

from .analyse import calc_comparison, plot_results


def get_mappings(index, file_hashes):
    patient_grouped_fields = dict()
    field_id_grouped_hashes = dict()

    for file_hash in file_hashes:
        delivery_details = index[file_hash]['delivery_details']
        patient_id = delivery_details['patient_id']
        field_id = delivery_details['field_id']

        if patient_id not in patient_grouped_fields:
            patient_grouped_fields[patient_id] = set()

        patient_grouped_fields[patient_id].add(field_id)

        if field_id not in field_id_grouped_hashes:
            field_id_grouped_hashes[field_id] = []

        field_id_grouped_hashes[field_id].append(file_hash)

    return patient_grouped_fields, field_id_grouped_hashes


def group_consecutive_logfiles(file_hashes, index):
    times = np.array([
        index[key]['local_time']
        for key in file_hashes
    ]).astype(np.datetime64)

    sort_reference = np.argsort(times)
    file_hashes = file_hashes[sort_reference]
    times = times[sort_reference]

    hours_4 = np.array(60 * 60 * 4).astype(np.timedelta64)
    split_locations = np.where(np.diff(times) >= hours_4)[0] + 1

    return np.split(file_hashes, split_locations)


def assert_array_agreement(unique_logfile_gantry_angles, mosaiq_gantry_angles,
                           allowed_deviation):
    difference_matrix = np.abs(
        unique_logfile_gantry_angles[:, None] - mosaiq_gantry_angles[None, :])
    agreement_matrix = difference_matrix <= allowed_deviation
    row_agreement = np.any(agreement_matrix, axis=1)
    at_least_one_agreement = np.all(row_agreement)

    assert at_least_one_agreement, (
        'There is a logfile gantry angle that deviates by more than {} degrees'
        ' from the Mosaiq control points. Unsure how to handle this.\n\n'
        'Logfile: {}\nMosaiq: {}\nDifference Matrix:\n{}\n'
        'Agreement Matrix:\n{}'.format(
            allowed_deviation, unique_logfile_gantry_angles,
            mosaiq_gantry_angles, difference_matrix, agreement_matrix))


def get_field_id_from_logfile_group(index, logfile_group):
    field_ids = []

    for logfile_hash in logfile_group:
        field_ids.append(index[logfile_hash]['delivery_details']['field_id'])

    assert len(np.unique(field_ids)) == 1

    field_id = field_ids[0]

    return field_id


def calc_normalisation(mosaiq_delivery_data):
    mosaiq_delivery_data.m
    all_gantry_angles = mosaiq_delivery_data.mudensity
    mosaiq_gantry_angles = np.unique(mosaiq_delivery_data.gantry)
    number_of_gantry_angles = len(mosaiq_gantry_angles)

    normalisation = np.sum(all_gantry_angles) / number_of_gantry_angles

    return normalisation


def calc_logfile_mu_density_bygantry(index, config, logfile_group,
                                     gantry_angle, grid_resolution=1):
    logfile_mu_density = None

    for filehash in logfile_group:
        filepath = get_filepath(index, config, filehash)
        logfile_delivery_data = DeliveryDatabases.from_logfile(filepath)

        a_logfile_mu_density = [
            get_grid(grid_resolution=grid_resolution),
            logfile_delivery_data.mudensity(
                gantry_angle, grid_resolution=grid_resolution)
        ]

        if logfile_mu_density is None:
            logfile_mu_density = a_logfile_mu_density
        else:
            assert np.all(logfile_mu_density[0] == a_logfile_mu_density[0])
            assert np.all(logfile_mu_density[1] == a_logfile_mu_density[1])
            logfile_mu_density[2] += a_logfile_mu_density[2]

    return logfile_mu_density


def compare_logfile_group_bygantry(index, config, cursor, logfile_group,
                                   gantry_angle, grid_resolution=1):
    field_id = get_field_id_from_logfile_group(index, logfile_group)

    mosaiq_delivery_data = multi_fetch_and_verify_mosaiq(cursor, field_id)

    mosaiq_mu_density = [
        get_grid(grid_resolution=grid_resolution),
        mosaiq_delivery_data.mudensity(
            gantry_angle, grid_resolution=grid_resolution)
    ]

    normalisation = calc_normalisation(mosaiq_delivery_data)

    logfile_mu_density = calc_logfile_mu_density_bygantry(
        index, config, logfile_group, gantry_angle)

    grid_xx = logfile_mu_density[0]
    grid_yy = logfile_mu_density[1]
    assert np.all(grid_xx == mosaiq_mu_density[0])
    assert np.all(grid_yy == mosaiq_mu_density[1])

    comparison = calc_comparison(
        logfile_mu_density[2], mosaiq_mu_density[2], normalisation)

    print(comparison)
    plot_results(
        grid_xx, grid_yy, logfile_mu_density[2], mosaiq_mu_density[2])

    return comparison


# -----------------------------------------------------------------------------
# TODO Refactor the following functions to be in terms of the above.
# -----------------------------------------------------------------------------


def get_logfile_delivery_data_bygantry(index, config, logfile_groups,
                                       mosaiq_gantry_angles):
    logfile_delivery_data_bygantry = dict()

    for logfile_group in logfile_groups:
        logfile_delivery_data_bygantry[logfile_group] = dict()

        for file_hash in logfile_group:
            filepath = get_filepath(index, config, file_hash)
            logfile_delivery_data = DeliveryDatabases.from_logfile(
                filepath)
            mu = np.array(logfile_delivery_data.monitor_units)

            filtered = logfile_delivery_data.filter_cps()

            mu = filtered.monitor_units
            mlc = filtered.mlc
            jaw = filtered.jaw
            logfile_gantry_angles = filtered.gantry

            gantry_tolerance = get_gantry_tolerance(index, file_hash, config)
            unique_logfile_gantry_angles = np.unique(logfile_gantry_angles)

            assert_array_agreement(
                unique_logfile_gantry_angles, mosaiq_gantry_angles,
                gantry_tolerance)

            logfile_delivery_data_bygantry[logfile_group][file_hash] = dict()

            for mosaiq_gantry_angle in mosaiq_gantry_angles:
                logfile_delivery_data_bygantry[logfile_group][file_hash][mosaiq_gantry_angle] = dict(
                )
                agrees_within_tolerance = (
                    np.abs(logfile_gantry_angles - mosaiq_gantry_angle) <= gantry_tolerance)

                logfile_delivery_data_bygantry[logfile_group][file_hash][
                    mosaiq_gantry_angle]['mu'] = mu[agrees_within_tolerance]
                logfile_delivery_data_bygantry[logfile_group][file_hash][
                    mosaiq_gantry_angle]['mlc'] = mlc[agrees_within_tolerance]
                logfile_delivery_data_bygantry[logfile_group][file_hash][
                    mosaiq_gantry_angle]['jaw'] = jaw[agrees_within_tolerance]

    return logfile_delivery_data_bygantry


def get_logfile_mu_density_bygantry(logfile_groups, mosaiq_gantry_angles,
                                    logfile_delivery_data_bygantry):
    logfile_mu_density_bygantry = dict()

    for logfile_group in logfile_groups:
        delivery_data = logfile_delivery_data_bygantry[logfile_group]
        logfile_mu_density_bygantry[logfile_group] = dict()

        for file_hash in logfile_group:
            for mosaiq_gantry_angle in mosaiq_gantry_angles:
                num_control_points = len(
                    delivery_data[file_hash][mosaiq_gantry_angle]['mu'])
                if num_control_points > 0:
                    mu_density = [
                        get_grid(),
                        delivery_data[file_hash][mosaiq_gantry_angle].mudensity()
                    ]

                    if mosaiq_gantry_angle not in logfile_mu_density_bygantry[logfile_group]:
                        logfile_mu_density_bygantry[logfile_group][mosaiq_gantry_angle] = list(
                            mu_density)
                    else:
                        assert np.all(
                            logfile_mu_density_bygantry[logfile_group][mosaiq_gantry_angle][0] == mu_density[0])
                        assert np.all(
                            logfile_mu_density_bygantry[logfile_group][mosaiq_gantry_angle][1] == mu_density[1])
                        logfile_mu_density_bygantry[logfile_group][mosaiq_gantry_angle][2] += mu_density[2]

    return logfile_mu_density_bygantry


def get_mosaiq_delivery_data_bygantry(mosaiq_delivery_data):
    mu = np.array(mosaiq_delivery_data.monitor_units)
    mlc = np.array(mosaiq_delivery_data.mlc)
    jaw = np.array(mosaiq_delivery_data.jaw)
    gantry_angles = np.array(mosaiq_delivery_data.gantry)
    unique_mosaiq_gantry_angles = np.unique(gantry_angles)

    mosaiq_delivery_data_bygantry = dict()

    for mosaiq_gantry_angle in unique_mosaiq_gantry_angles:
        gantry_angle_matches = gantry_angles == mosaiq_gantry_angle

        diff_mu = np.concatenate([[0], np.diff(mu)])[gantry_angle_matches]
        gantry_angle_specific_mu = np.cumsum(diff_mu)

        mosaiq_delivery_data_bygantry[mosaiq_gantry_angle] = dict()
        mosaiq_delivery_data_bygantry[mosaiq_gantry_angle]['mu'] = gantry_angle_specific_mu
        mosaiq_delivery_data_bygantry[mosaiq_gantry_angle]['mlc'] = mlc[gantry_angle_matches]
        mosaiq_delivery_data_bygantry[mosaiq_gantry_angle]['jaw'] = jaw[gantry_angle_matches]

    return mosaiq_delivery_data_bygantry


def get_mosaiq_mu_density_bygantry(mosaiq_delivery_data_bygantry):
    mosaiq_mu_density_bygantry = dict()
    mosaiq_gantry_angles = mosaiq_delivery_data_bygantry.keys()

    for mosaiq_gantry_angle in mosaiq_gantry_angles:
        mu_density = [
            get_grid(),
            mosaiq_delivery_data_bygantry[mosaiq_gantry_angle].mudensity()
        ]
        mosaiq_mu_density_bygantry[mosaiq_gantry_angle] = mu_density

    return mosaiq_mu_density_bygantry


def get_comparison_results(mosaiq_mu_density_bygantry,
                           logfile_mu_density_bygantry, normalisation):
    comparison_results = dict()
    mosaiq_gantry_angles = mosaiq_mu_density_bygantry.keys()
    logfile_groups = list(logfile_mu_density_bygantry.keys())

    for mosaiq_gantry_angle in mosaiq_gantry_angles:
        comparison_results[mosaiq_gantry_angle] = dict()
        comparison_results[mosaiq_gantry_angle]['comparisons'] = {}

        grid_xx = mosaiq_mu_density_bygantry[mosaiq_gantry_angle][0]
        grid_yy = mosaiq_mu_density_bygantry[mosaiq_gantry_angle][1]
        mosaiq_mu_density = mosaiq_mu_density_bygantry[mosaiq_gantry_angle][2]

        for logfile_group in logfile_groups:
            assert np.all(
                grid_xx == logfile_mu_density_bygantry[logfile_group][mosaiq_gantry_angle][0])
            assert np.all(
                grid_yy == logfile_mu_density_bygantry[logfile_group][mosaiq_gantry_angle][1])

            logfile_mu_density = logfile_mu_density_bygantry[logfile_group][mosaiq_gantry_angle][2]

            comparison = calc_comparison(
                logfile_mu_density, mosaiq_mu_density, normalisation)
            comparison_results[mosaiq_gantry_angle]['comparisons'][logfile_group] = comparison

        comparisons = np.array([
            comparison_results[mosaiq_gantry_angle]['comparisons'][logfile_group]
            for logfile_group in logfile_groups
        ])

        comparison_results[mosaiq_gantry_angle]['median'] = np.median(
            comparisons)
        ref = np.argmin(np.abs(
            comparisons -
            comparison_results[mosaiq_gantry_angle]['median']
        ))
        comparison_results[mosaiq_gantry_angle]['median_filehash_group'] = logfile_groups[ref]
        comparison_results[mosaiq_gantry_angle]['filehash_groups'] = logfile_groups

    return comparison_results


def get_mu_densities_for_file_hashes(index, config, cursor, file_hashes):
    field_ids = {
        index[file_hash]['delivery_details']['field_id']
        for file_hash in file_hashes
    }

    assert len(field_ids) == 1
    field_id = field_ids.pop()

    logfile_groups = group_consecutive_logfiles(file_hashes, index)
    logfile_groups = [
        tuple(group)
        for group in logfile_groups
    ]

    mosaiq_delivery_data = multi_fetch_and_verify_mosaiq(cursor, field_id)
    mosaiq_gantry_angles = np.unique(mosaiq_delivery_data.gantry)

    logfile_delivery_data_bygantry = get_logfile_delivery_data_bygantry(
        index, config, logfile_groups, mosaiq_gantry_angles)
    logfile_mu_density_bygantry = get_logfile_mu_density_bygantry(
        logfile_groups, mosaiq_gantry_angles, logfile_delivery_data_bygantry)
    mosaiq_delivery_data_bygantry = get_mosaiq_delivery_data_bygantry(
        mosaiq_delivery_data)
    mosaiq_mu_density_bygantry = get_mosaiq_mu_density_bygantry(
        mosaiq_delivery_data_bygantry)

    normalisation = calc_normalisation(mosaiq_delivery_data)

    return (
        mosaiq_mu_density_bygantry, logfile_mu_density_bygantry,
        normalisation)


def get_mu_densities_for_field_id(index, config, cursor, field_id, field_id_grouped_hashes):
    file_hashes = np.array(field_id_grouped_hashes[field_id])
    mu_densities = get_mu_densities_for_file_hashes(
        index, config, cursor, file_hashes)

    return mu_densities


def get_comparisons_byfield(index, config, cursor, field_ids,
                            field_id_grouped_hashes):
    comparisons_byfield = dict()

    for field_id in field_ids:
        mu_densities = get_mu_densities_for_field_id(
            index, config, cursor, field_id, field_id_grouped_hashes)
        comparison_results = get_comparison_results(*mu_densities)

        comparisons_byfield[field_id] = comparison_results

    return comparisons_byfield



install_requires = [
    "attrs",
    "matplotlib",
    "numpy >= 1.12",
    "pandas",
    "pymedphys_databases >= 0.10.0, < 0.11.0",
    "pymedphys_fileformats >= 0.10.0, < 0.11.0",
    "pymedphys_mudensity >= 0.10.0, < 0.11.0",
    "pymedphys_utilities >= 0.10.0, < 0.11.0",
    "python_dateutil"
]
